"""
The command-line entry point into the hera-py system.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: March 2019
"""
import functools
import sys
from typing import List, Optional

from .assembler import assemble_and_print
from .data import HERAError, Settings, VOLUME_QUIET, VOLUME_VERBOSE
from .debugger import debug
from .loader import load_program_from_file
from .op import disassemble
from .utils import format_int, read_file_or_stdin
from .vm import VirtualMachine


def external_main(argv=None) -> None:
    """
    A wrapper around main that ignores its return value, so it is not printed to the
    console when the program exits.
    """
    main(argv)


def main(argv=None) -> Optional[VirtualMachine]:
    """The main entry point into hera-py."""
    settings = parse_args(argv)
    path = settings.path

    if not sys.stderr.isatty():
        settings.color = False

    if settings.mode == "preprocess":
        settings.allow_interrupts = True
        main_preprocess(path, settings)
        return None
    elif settings.mode == "debug":
        main_debug(path, settings)
        return None
    elif settings.mode == "assemble":
        settings.allow_interrupts = True
        main_assemble(path, settings)
        return None
    elif settings.mode == "disassemble":
        main_disassemble(path, settings)
        return None
    else:
        return main_execute(path, settings)


def main_debug(path: str, settings: Settings) -> None:
    """Debug the program."""
    program = load_program_from_file(path, settings)
    debug(program, settings)


def main_execute(path: str, settings: Settings) -> VirtualMachine:
    """Execute the program."""
    program = load_program_from_file(path, settings)

    vm = VirtualMachine(settings)
    vm.run(program)

    # Check if the program exited prematurely due to throttling.
    if settings.throttle is not False and vm.op_count == settings.throttle:
        sys.stderr.write(
            "Program throttled after {} instructions.\n".format(vm.op_count)
        )

    settings.warning_count += vm.warning_count

    if settings.volume != VOLUME_QUIET:
        dump_state(vm, settings)

    return vm


def main_preprocess(path: str, settings: Settings) -> None:
    """Preprocess the program and print it to stdout."""
    program = load_program_from_file(path, settings)

    if not settings.obfuscate:
        if program.data:
            print("[DATA]")
            for data_op in program.data:
                print("  {}".format(data_op))

            if program.code:
                print("\n[CODE]")

        for i, op in enumerate(program.code):
            print("  {:0>4}  {}".format(i, op))
    else:
        for data_op in program.data:
            print("{}".format(data_op))

        for i, op in enumerate(program.code):
            assembled = op.assemble()
            v = (assembled[0] << 8) + assembled[1]
            print("OPCODE(0x{:x})".format(v))


def main_assemble(path: str, settings: Settings) -> None:
    """Assemble the program into machine code and print the hex output to stdout."""
    program = load_program_from_file(path, settings)
    assemble_and_print(program, settings)


def main_disassemble(path: str, settings: Settings) -> None:
    """
    Disassemble the machine code (expressed as newline-separated hex numbers, without
    the "0x" prefix), and print the HERA output to stdout.
    """
    text = read_file_or_stdin(path, settings)
    for line in text.splitlines():
        try:
            v = int(line, base=16)
        except ValueError:
            print("// Invalid hex literal: {}".format(line))
            continue

        try:
            print(disassemble(v))
        except HERAError:
            print("// Unknown instruction: {}".format(line))


def parse_args(argv: List[str]) -> Settings:
    if argv is None:
        argv = sys.argv[1:]

    flags = {}
    posargs = []
    after_flags = False
    i = 0
    while i < len(argv):
        arg = argv[i]
        longarg = short_to_long(arg)
        if longarg == "--":
            after_flags = True
        elif longarg in FLAGS:
            # --throttle is the only flag that takes an argument.
            if longarg == "--throttle":
                if i == len(argv) - 1 or not argv[i + 1].isdigit():
                    sys.stderr.write("--throttle takes one integer argument.\n")
                    sys.exit(1)
                flags[longarg] = int(argv[i + 1])
                i += 1
            else:
                flags[longarg] = True
        elif not after_flags and longarg.startswith("-") and len(longarg) > 1:
            sys.stderr.write("Unrecognized flag: " + arg + "\n")
            sys.exit(1)
        else:
            posargs.append(longarg)
        i += 1

    if "--help" in flags:
        if len(flags) == 1 and not posargs:
            print(HELP)
            sys.exit(0)
        else:
            sys.stderr.write(
                "--help may not be combined with other flags or commands.\n"
            )
            sys.exit(1)

    if "--version" in flags:
        if len(flags) == 1 and not posargs:
            print(VERSION)
            sys.exit(0)
        else:
            sys.stderr.write(
                "--version may not be combined with other flags or commands.\n"
            )
            sys.exit(1)

    if "--credits" in flags:
        if len(flags) == 1 and not posargs:
            print(CREDITS)
            sys.exit(0)
        else:
            sys.stderr.write(
                "--credits may not be combined with other flags or commands.\n"
            )
            sys.exit(1)

    if len(posargs) == 0:
        sys.stderr.write("No file path supplied.\n")
        sys.exit(1)
    elif len(posargs) > 1:
        sys.stderr.write("Too many file paths supplied.\n")
        sys.exit(1)

    if "debug" in flags:
        mode = "debug"
    elif "assemble" in flags:
        mode = "assemble"
    elif "preprocess" in flags:
        mode = "preprocess"
    elif "disassemble" in flags:
        mode = "disassemble"
    else:
        mode = ""

    for picky_flag, valid_modes in PICKY_FLAGS.items():
        if picky_flag in flags:
            if mode not in valid_modes:
                sys.stderr.write(
                    "{} is not compatible with the chosen mode.\n".format(picky_flag)
                )
                sys.exit(1)

    if "--quiet" in flags and "--verbose" in flags:
        sys.stderr.write("--quiet and --verbose are incompatible.\n")
        sys.exit(1)

    for f in FLAGS:
        if f not in flags:
            flags[f] = False

    settings = Settings()
    settings.path = posargs[0]
    settings.mode = mode

    settings.allow_interrupts = settings.mode in ("assemble", "preprocess")
    settings.code = flags["--code"]
    settings.color = not flags["--no-color"]
    settings.data = flags["--data"]
    if flags["--big-stack"]:
        # Arbitrary value copied over from HERA-C.
        settings.data_start = 0xC167
    settings.no_debug_ops = flags["--no-debug-ops"]
    settings.obfuscate = flags["--obfuscate"]
    settings.stdout = flags["--stdout"]
    settings.throttle = flags["--throttle"]
    settings.warn_octal_on = not flags["--warn-octal-off"]
    settings.warn_return_on = not flags["--warn-return-off"]
    if flags["--verbose"]:
        settings.volume = VOLUME_VERBOSE
    elif flags["--quiet"]:
        settings.volume = VOLUME_QUIET

    return settings


def short_to_long(arg: str) -> str:
    if arg == "-h":
        return "--help"
    elif arg == "-v":
        return "--version"
    elif arg == "-q":
        return "--quiet"
    else:
        return arg


def dump_state(vm: VirtualMachine, settings: Settings) -> None:
    """Print the state of the virtual machine to standard output."""
    # Make sure that all program output has been printed.
    sys.stdout.flush()

    # Redefine print in this function to use stderr.
    nprint = functools.partial(print, file=sys.stderr)

    verbose = settings.volume == VOLUME_VERBOSE
    if verbose:
        last_register = 15
    else:
        last_register = 10
        while last_register > 0 and vm.registers[last_register] == 0:
            last_register -= 1

    nprint("\nVirtual machine state after execution:")
    for i, value in enumerate(vm.registers[1 : last_register + 1], start=1):
        rname = "    R" + str(i) + (" " if i < 10 else "")
        nprint("{} = {}".format(rname, format_int(value)))

    if last_register > 0:
        nprint()
    else:
        nprint("    R1 through R10 are all zero.\n")

    flags = [
        vm.flag_carry_block,
        vm.flag_carry,
        vm.flag_overflow,
        vm.flag_zero,
        vm.flag_sign,
    ]
    if not verbose and all(flags):
        nprint("    All flags are ON")
    elif not verbose and all(not f for f in flags):
        nprint("    All flags are OFF")
    else:
        nprint("    Carry-block flag is " + ("ON" if vm.flag_carry_block else "OFF"))
        nprint("    Carry flag is " + ("ON" if vm.flag_carry else "OFF"))
        nprint("    Overflow flag is " + ("ON" if vm.flag_overflow else "OFF"))
        nprint("    Zero flag is " + ("ON" if vm.flag_zero else "OFF"))
        nprint("    Sign flag is " + ("ON" if vm.flag_sign else "OFF"))

    if settings.warning_count > 0:
        c = settings.warning_count
        nprint("\n{} warning{} emitted.".format(c, "" if c == 1 else "s"))


FLAGS = {
    "--big-stack",
    "--code",
    "--credits",
    "--data",
    "--help",
    "--no-color",
    "--no-debug-ops",
    "--obfuscate",
    "--quiet",
    "--stdout",
    "--throttle",
    "--verbose",
    "--version",
    "--warn-octal-off",
    "--warn-return-off",
    "assemble",
    "debug",
    "disassemble",
    "preprocess",
}

# Map from flag names to compatible modes, e.g. "--big-stack" is only compatible with
# the run, debug and assemble modes.
PICKY_FLAGS = {
    "--big-stack": ["", "debug", "assemble"],
    "--obfuscate": ["preprocess"],
    "--throttle": [""],
    "--warn-return-off": ["", "debug"],
    "--code": ["assemble"],
    "--data": ["assemble"],
    "--stdout": ["assemble"],
}

VERSION = "hera-py 0.8.1 for HERA version 2.4"
CREDITS = (
    VERSION
    + """

Developed by Ian Fisher (HC '19) - iafisher@protonmail.com."""
)
HELP = """\
hera: an interpreter for the Haverford Educational RISC Architecture.

Usage:
    hera <path>
    hera debug <path>
    hera assemble <path>
    hera preprocess <path>
    hera disassemble <path>

Common options:
    -h, --help         Show this message and exit.
    -v, --version      Show the version and exit.
    --credits          Print the credits for hera-py development.

    --no-color         Do not print colored output.
    --no-debug-ops     Disallow debugging instructions.
    -q --quiet         Set output level to quiet.
    --verbose          Set output level to verbose.
    --warn-octal-off   Do not print warnings for zero-prefixed integer literals.

Interpreter and debugger options:
    --big-stack        Reserve more space for the stack.
    --throttle <n>     Exit after <n> instructions have been executed.
    --warn-return-off  Do not print warnings for invalid RETURN addresses.

Assembler options:
    --code             Only output the assembled code.
    --data             Only output the assembled data.
    --stdout           Print the assembled program to stdout instead of creating
                       files.
"""



"""
The standard library for the Tiger programming language, implemented in HERA.

All of the pure HERA functions in this module have been copied from the original
Tiger standard library file for HERA-C, written by Dave Wonnacott.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: February 2019
"""
import sys

from .utils import from_u16, to_u16


def tiger_printint_stack(vm):
    print(from_u16(vm.load_memory(vm.registers[14] + 3)), end="")


def tiger_printbool_stack(vm):
    v = vm.load_memory(vm.registers[14] + 3)
    print("false" if v == 0 else "true", end="")


def tiger_print_stack(vm):
    addr = vm.load_memory(vm.registers[14] + 3)
    n = vm.load_memory(addr)
    for i in range(n):
        print(chr(vm.load_memory(addr + i + 1)), end="")


def tiger_println_stack(vm):
    addr = vm.load_memory(vm.registers[14] + 3)
    n = vm.load_memory(addr)
    for i in range(n):
        print(chr(vm.load_memory(addr + i + 1)), end="")
    print()


def tiger_div_stack(vm):
    left = vm.load_memory(vm.registers[14] + 3)
    right = vm.load_memory(vm.registers[14] + 4)
    result = left // right if right != 0 else 0
    vm.store_memory(vm.registers[14] + 3, result)


def tiger_mod_stack(vm):
    left = vm.load_memory(vm.registers[14] + 3)
    right = vm.load_memory(vm.registers[14] + 4)
    result = left % right if right != 0 else 0
    vm.store_memory(vm.registers[14] + 3, result)


def tiger_getchar_ord_stack(vm):
    if vm.input_pos >= len(vm.input_buffer):
        vm.readline()

    if len(vm.input_buffer) > 0:
        c = ord(vm.input_buffer[vm.input_pos])
        vm.input_pos += 1
    else:
        c = 0
    vm.store_memory(vm.registers[14] + 3, c)


def tiger_ungetchar_stack(vm):
    vm.input_pos -= 1


def tiger_putchar_ord_stack(vm):
    print(chr(vm.load_memory(vm.registers[14] + 3)), end="")


def tiger_flush_stack(vm):
    sys.stdout.flush()


def tiger_getline_preamble_stack(vm):
    vm.readline()
    vm.registers[1] = len(vm.input_buffer) + 1


def tiger_getline_epilogue_stack(vm):
    addr = vm.registers[1]
    vm.store_memory(addr, len(vm.input_buffer))
    for i, c in enumerate(vm.input_buffer, start=1):
        vm.store_memory(addr + i, ord(c))


# The standard library with parameters-on-the-stack functions.
TIGER_STDLIB_STACK = """
LABEL(printint)
  __eval("stdlib.tiger_printint_stack(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(print)
  __eval("stdlib.tiger_print_stack(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(println)
  __eval("stdlib.tiger_println_stack(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(div)
  __eval("stdlib.tiger_div_stack(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(mod)
  __eval("stdlib.tiger_mod_stack(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(getchar_ord)
  __eval("stdlib.tiger_getchar_ord_stack(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(putchar_ord)
  __eval("stdlib.tiger_putchar_ord_stack(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(flush)
  __eval("stdlib.tiger_flush_stack(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(printbool)
  __eval("stdlib.tiger_printbool_stack(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(ungetchar)
  __eval("stdlib.tiger_ungetchar_stack(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(getline)
  __eval("stdlib.tiger_getline_preamble_stack(vm)")
  MOVE(R12, SP)
  INC(SP, 4)

  // R1 was set to the length of the string by tiger_getline_preamble_stack
  STORE(R1, 4, R12)

  CALL(R12, malloc)
  LOAD(R1, 3, R12)
  DEC(SP, 4)

  __eval("stdlib.tiger_getline_epilogue_stack(vm)")


LABEL(exit)
  HALT()


LABEL(size)
     INC(SP,1)
     STORE(R1,4,FP)
     LOAD(R1,3,FP) // Reg 1 <-- argument (address of string)
     LOAD(R1,0,R1)  // Reg 1 now has the size of string
     STORE(R1,3,FP) // put it (the size) into return area
     LOAD(R1,4,FP)
     DEC(SP,1)
     RETURN(FP_alt, PC_ret)


LABEL(concat)
// vars & reg: s1ptr/s2ptr(1), nptr(2), n(3), tmp(4), tmp2(5)

// store needed registers
     STORE(PC_ret, 0,FP)
     STORE(FP_alt,1,FP)
     INC(SP,9) // space for registers and calling functions

     STORE(R1,5,FP)
     STORE(R2,6,FP)
     STORE(R3,7,FP)
     STORE(R4,8,FP)
     STORE(R5,9,FP)

     MOVE(FP_alt,SP)

// malloc length(s1)+length(s2)+1
     LOAD(R3,3,FP)
     LOAD(R3,0,R3)	// get s1's length
     LOAD(R2,4,FP)
     LOAD(R2,0,R2)	// get s2's length
     ADD(R3,R2,R3)
     INC(R3,1)
     STORE(R3, 3,FP_alt)
     CALL(FP_alt,malloc)     // get space

     LOAD(R2, 3,FP_alt)     // allocated space
     DEC(R3,1)   	// sum of the lengths again
     STORE(R3, 0,R2)    // Set result's size
     LOAD(R1, 3,FP)	// s1
     STORE(R2, 3,FP)    // SAVE RESULT, AND KILL S1
     INC(R2,1)  	// R2 now next available space

     LOAD(R3, 0,R1)	// s1's size
     INC(R1,1)  	// sptr is now s1's initial character

// memcpy(nptr,s1ptr,length(s1))
     CALL(FP_alt,tstdlib_label_local_memcpy_reg)   // memcpy(sptr,nptr,n)
// memcpy(nptr,s2ptr,length(s2))
     LOAD(R1,4,FP)
     LOAD(R3,0,R1)	// get s2's length
     INC(R1,1)		// sptr is now s2's initial character
     CALL(FP_alt,tstdlib_label_local_memcpy_reg)   // memcpy(sptr,nptr,n)

     LOAD(R1,5,FP)
     LOAD(R2,6,FP)
     LOAD(R3,7,FP)
     LOAD(R4,8,FP)
     LOAD(R5,9,FP)
     LOAD(PC_ret, 0,FP)
     LOAD(FP_alt,1,FP)
     DEC(SP, 9)
     RETURN(FP_alt, PC_ret)		// value saved a while ago


// ** tstrcmp(s1:string, s2:string) : int
LABEL(tstrcmp)  // tiger strcmp -- like C strcmp
		// tstrcmp(a,b) returns neg # if a < b, 0 if =, pos # if a > b
		//  Note that this is also similar to doing CMP(a,b)
// Get some space, save some registers
     STORE(PC_ret, 0,FP) // For good practice
     STORE(FP_alt,1,FP)
     INC(SP,6)
     STORE(R1,10,FP)
     STORE(R2,5,FP)
     STORE(R3,6,FP)
     STORE(R4,7,FP)
     STORE(R5,8,FP)
     STORE(R6,9,FP)

// r3 = min(size(a), size(b))
     LOAD(R1,3,FP)	// r1 points to size of a
     LOAD(R2,4,FP)
     LOAD(R5,0,R1)		// r5 has size of a
     LOAD(R6,0,R2)		// r6 has size of b
     CMP(R5,R6)
     BLR(tstdlib_label_b_was_longer)
     ADD(R3,R5,R0)	// r3 = r5
     BR(tstdlib_label_got_min)
     LABEL(tstdlib_label_b_was_longer)
     ADD(R3,R6,R0)	// r3 = r6
     LABEL(tstdlib_label_got_min)  // now r3 is min. size
// r1 = address of 1st "real" char of a,
     INC(R1,1)
// r2 = address of 1st "real" char of b
// could also use this first:  LOAD(R2,4,FP)
     INC(R2,1)
// r4 = 0
     ADD(R4,R0,R0)
// while(r4<r3) {
     LABEL(tstdlib_label_start_of_while_in_tstrcmp)
     CMP(R4,R3)
     BGER(tstdlib_label_get_out_of_while_in_tstrcmp)
//  if (a[r4] < b[r4])  // add *(r4 + a) < *(r4 + b)
     ADD(R5,R4,R1)  // r5 = r4+a
     LOAD(R5,0,R5) // r5 = *(r4+a)
     ADD(R6,R4,R2)
     LOAD(R6,0,R6)
     CMP(R5,R6)
     BGER(tstdlib_label_not_a_smaller)
//  	return -1
     SET(R5,-1)
     STORE(R5,3,FP)
     BR(tstdlib_label_do_return)
     LABEL(tstdlib_label_not_a_smaller)
//  if (a[r4] > b[r4])
     CMP(R6,R5)
     BGER(tstdlib_label_not_b_smaller)
//		return 1
     SET(R5,1)
     STORE(R5,3,FP)
     BR(tstdlib_label_do_return)
     LABEL(tstdlib_label_not_b_smaller)
// 	r4++;
     INC(R4,1)
// }
     BR(tstdlib_label_start_of_while_in_tstrcmp)
     LABEL(tstdlib_label_get_out_of_while_in_tstrcmp)
// return size(a) - size(b)
     LOAD(R1,3,FP)	// r1 points to size of a
     LOAD(R2,4,FP)
     LOAD(R5,0,R1)		// r5 has size of a
     LOAD(R6,0,R2)		// r6 has size of b
     SUB(R5,R5,R6)		// r5 has size of a - size of b
     STORE(R5,3,FP)
     LABEL(tstdlib_label_do_return)
     LOAD(R1,10,FP)
     LOAD(R2,5,FP)
     LOAD(R3,6,FP)
     LOAD(R4,7,FP)
     LOAD(R5,8,FP)
     LOAD(R6,9,FP)
     LOAD(PC_ret, 0,FP)
     LOAD(FP_alt,1,FP)
     DEC(SP,6)
     RETURN(FP_alt, PC_ret)


//   substring(s:string, first:int, n:int) : string
LABEL(substring)
// vars & reg: sptr(1), nptr(2), n(3), tmp(4)
// store needed registers
     STORE(PC_ret, 0,FP)
     STORE(FP_alt,1,FP)
     INC(SP,9) // Space for registers and all function calls done by substring, in one fell swoop
     STORE(R1,6,FP)
     STORE(R2,7,FP)
     STORE(R3,8,FP)
     STORE(R4,9,FP)

     MOVE(FP_alt,SP)

// first check parameters
     LOAD(R1,3,FP)	// address of s  (points to its size)
     LOAD(R3,5,FP)      // R3 now "n" from here on

     LOAD(R4,4,FP)	// "first"
     ADD(R0,R4,R0)
     BS(substring_bad_params)   // check for first <0

     ADD(R0,R3,R0)
     BS(substring_bad_params)   // check for n<0

     ADD(R4,R4,R3)      // first+n
     LOAD(R2,0,R1)      // size of s
     CMP(R2,R4)
     BL(substring_bad_params)  // check for size(s)<first+n

// nptr = next free address   (allocate n+1 cells, save address for later)
     ADD(R4,R3,R0)      // R4 = n
     INC(R4,1)
     STORE(R4,3,FP_alt)
     CALL(FP_alt,malloc)
     LOAD(R2,3,FP_alt)	// R2 now nptr from here on

// save result of malloc (nptr) as return value   AND KILL "s"
     STORE(R2,3,FP)

// sptr = &s[first]
     INC(R1,1)  	// R1 is sptr from here on
     LOAD(R4,4,FP)	// "first"
     ADD(R1,R4,R1) 	// sptr now points to 1st char to grab
// *(nptr++) = n
     STORE(R3,0,R2)
     INC(R2,1)
     CALL(FP_alt,tstdlib_label_local_memcpy_reg)   // memcpy(sptr,nptr,n)
     LOAD(R1,6,FP)
     LOAD(R2,7,FP)
     LOAD(R3,8,FP)
     LOAD(R4,9,FP)
     LOAD(PC_ret, 0,FP)
     LOAD(FP_alt,1,FP)
     DEC(SP, 9)
     RETURN(FP_alt, PC_ret)		// value saved a while ago

     LABEL(substring_bad_params)
     SET(R1,substring_got_bad_params)
     STORE(R1,3,FP_alt)
     CALL(FP_alt,print)
     CALL(FP_alt,exit)


LABEL(malloc)
     STORE(PC_ret, 0,FP)
     STORE(FP_alt,1,FP)
     INC(SP,3)
     STORE(R1,4,FP)
     STORE(R2,5,FP)
     STORE(R3,6,FP)

     LOAD(R1,3,FP) // Reg 1 <-- argument (number of cells needed)
     SET(R3,first_space_for_fsheap)
     LOAD(R2,0,R3) // R2 is address of first free
   BNZ(fsheap_already_initialized)
     MOVE(R2, R3)
     INC(R2, 1)    // initialize fs heap to its first free element then allocate from there
     STORE(R2,0,R3) // defensive programming --- this should be overwritten later and thus dead
   LABEL(fsheap_already_initialized)
     STORE(R2,3,FP)// RETURN VALUE DEFINED HERE

     ADD(R2,R2,R1) // ALLOCATE -- FAIL BELOW IF WRAPPED PAST END OF ADDRESS SPACE OR "last_space_for_heap"
     BC(malloc_give_up)
     SET(R1,last_space_for_fsheap)
     CMP(R2, R1)   // want (unsigned) R2 < R1, so R2-R1 should BORROW, i.e. NOT carry!
     BC(malloc_give_up)
     STORE(R2,0,R3)   // STORE BACK THE NEXT FREE SPACE

     LOAD(R1,4,FP)
     LOAD(R2,5,FP)
     LOAD(R3,6,FP)
     LOAD(PC_ret, 0,FP)
     LOAD(FP_alt,1,FP)
     DEC(SP, 3)
     RETURN(FP_alt, PC_ret)	// Normal return from malloc

   LABEL(malloc_inconsistent)
     SET(R1, malloc_inconsistent_error)
     BR(malloc_exit)
   LABEL(malloc_give_up)
     SET(R1,malloc_out_of_memory_error)
   LABEL(malloc_exit)

    MOVE(FP_alt,SP)
    INC(SP,4)

     STORE(R1,3,FP_alt)
     CALL(FP_alt,print)
     CALL(FP_alt,exit)
    DEC(SP, 4) // just to match

LABEL(tstdlib_label_local_memcpy_reg)
     // move n(reg 3) bytes from location sptr(reg 1) to location nptr (reg 2)
     // modifies registers 1, 2, 3, and 4
     //  (after end, r2 and r1 have been increased by original r3)
     ADD(R4,Rt,R0)
// while(n>0)
     LABEL(tstdlib_label_memcpy_while_begin)
     OR(R0,R0,R3)  // set flags for R3
     BZ(tstdlib_label_memcpy_while_end)
//   *(nptr++) = *(sptr++)
     LOAD(Rt,0,R1)
     INC(R1,1)
     STORE(Rt,0,R2)
     INC(R2,1)
//   n--
     DEC(R3,1)
     BR(tstdlib_label_memcpy_while_begin)
     LABEL(tstdlib_label_memcpy_while_end)
     ADD(Rt,R4,R0)
     RETURN(FP_alt, PC_ret)


//   ord(s:string) : int
LABEL(ord)
     INC(SP,1)
     STORE(R1,4,FP)
     LOAD(R1,3,FP) // Reg 1 <-- argument (address of string)
     LOAD(R1,1,R1)  // Reg 1 now has the 1st character of the string
     STORE(R1,3,FP) // put it (the character) into return area
     LOAD(R1,4,FP)
     DEC(SP, 1)
     RETURN(FP_alt, PC_ret)


 //   chr(i:int) : string
LABEL(chr)
     STORE(PC_ret, 0,FP)
     STORE(FP_alt,1,FP)
     INC(SP,2)
     STORE(R1,4,FP)
     STORE(R2,5,FP)
     MOVE(FP_alt,SP) // malloc(2)
     INC(SP,5)
     SET(r1,2)
     STORE(r1,3,FP_alt)
     CALL(FP_alt,malloc)
     LOAD(r2,3,FP_alt)  // Reg 2 <-- result of malloc
     DEC(SP, 5)
     SET(R1,1)
     STORE(R1,0,R2)	// set string length
     LOAD(R1,3,FP)	// Reg 1 <-- argument (the integer value)
     STORE(R1,1,R2)	// set string's one character
     STORE(R2,3,FP)	// address of string into return area
     LOAD(R1,4,FP)
     LOAD(R2,5,FP)
     LOAD(PC_ret, 0,FP)
     LOAD(FP_alt,1,FP)
     DEC(SP,2)
     RETURN(FP_alt, PC_ret)


//    not(i:int) : int
LABEL(not)
     STORE(PC_ret, 0,FP)
     STORE(FP_alt,1,FP)
     INC(SP,1)
     STORE(R1,4,FP)
     LOAD(R1,3,FP) // Reg 1 <-- argument
     CMP(R1,R0)
     BZ(tstdlib_label_arg_was_false)
     SET(R1,0)
     STORE(R1,3,FP) // put false into return area
     BR(tstdlib_label_return_from_not)
     LABEL(tstdlib_label_arg_was_false)
     SET(R1,1)
     STORE(R1,3,FP) // put true into return area
     LABEL(tstdlib_label_return_from_not)
     LOAD(R1,4,FP)
     LOAD(PC_ret, 0,FP)
     LOAD(FP_alt,1,FP)
     DEC(SP,1)
     RETURN(FP_alt, PC_ret)


//   getchar() : string
LABEL(getchar)
    STORE(PC_ret, 0,FP)
    STORE(FP_alt,1,FP)
    INC(SP, 7)  // save registers, set up for call
    STORE(R1, 4,FP)
    STORE(R2, 5,FP)

    MOVE(FP_alt,SP)

    SET(R1, 2)
    STORE(R1, 3,FP_alt)	// Added this Oct 2011 based on instinct that it needs to be here...
    CALL(FP_alt, malloc)
    LOAD(R2, 3,FP_alt)
    STORE(R2, 3,FP)     // this is the string we'll return
    SETLO(R1, 1)	// Size of string we're reading
    STORE(R1, 0,R2)
    CALL(FP_alt, getchar_ord)
    LOAD(R1, 3,FP_alt)	// R1 is now the character
    STORE(R1, 1,R2)
    LOAD(R2, 5,FP)
    LOAD(R1, 4,FP)
    LOAD(PC_ret, 0,FP)
    LOAD(FP_alt,1,FP)
    DEC(SP, 7)
    RETURN(FP_alt, PC_ret)
"""


# The data segment for the parameters-on-the-stack functions.
TIGER_STDLIB_STACK_DATA = """
CONSTANT(first_space_for_fsheap, 0x4000)
CONSTANT(last_space_for_fsheap, 0xbfff)


DLABEL(tiger_stdlib_endl)  /* manually build an "end-of-line" string */
INTEGER(1)
INTEGER(10)  /* control-j -- an end-of-line */

DLABEL(tiger_stdlib_hex_prefix)
TIGER_STRING("0x")
DLABEL(tiger_stdlib_printing_one_char_tmp)
TIGER_STRING(" ")
DLABEL(tiger_stdlib_ungetchar_one_char_tmp)  // size = defined; char = the char
INTEGER(0)  // 0=undefined; 1=defined
INTEGER(0)  // the character

DLABEL(malloc_inconsistent_error)
TIGER_STRING("internal inconsistency in malloc -- program terminated\\n")

DLABEL(malloc_out_of_memory_error)
TIGER_STRING("out of memory in malloc -- program terminated\\n")

DLABEL(substring_got_bad_params)
TIGER_STRING("bad parameters to substring -- program terminated\\n")

DLABEL(tstdlib_not_implemented)
TIGER_STRING("this function is not yet implemented in tiger standard libarry; halting\\n")
DLABEL(tstdlib_not_tested)
TIGER_STRING("WARNING: Entering untested territory in tiger stdlib\\n")
"""


def tiger_printint_reg(vm):
    print(from_u16(vm.registers[1]), end="")


def tiger_printbool_reg(vm):
    print("false" if vm.registers[1] == 0 else "true", end="")


def tiger_print_reg(vm):
    addr = vm.registers[1]
    n = vm.load_memory(addr)
    for i in range(n):
        print(chr(vm.load_memory(addr + i + 1)), end="")


def tiger_println_reg(vm):
    addr = vm.registers[1]
    n = vm.load_memory(addr)
    for i in range(n):
        print(chr(vm.load_memory(addr + i + 1)), end="")
    print()


def tiger_div_reg(vm):
    left = vm.registers[1]
    right = vm.registers[2]
    vm.registers[1] = left // right if right != 0 else 0


def tiger_mod_reg(vm):
    left = vm.registers[1]
    right = vm.registers[2]
    vm.registers[1] = left % right if right != 0 else 0


def tiger_getchar_ord_reg(vm):
    if vm.input_pos >= len(vm.input_buffer):
        vm.readline()

    if len(vm.input_buffer) > 0:
        c = ord(vm.input_buffer[vm.input_pos])
        vm.input_pos += 1
    else:
        c = 0
    vm.registers[1] = c


def tiger_ungetchar_reg(vm):
    vm.input_pos -= 1


def tiger_putchar_ord_reg(vm):
    print(chr(vm.registers[1]), end="")


def tiger_flush_reg(vm):
    sys.stdout.flush()


def tiger_getline_preamble_reg(vm):
    vm.readline()
    vm.registers[1] = len(vm.input_buffer) + 1


def tiger_getline_epilogue_reg(vm):
    addr = vm.registers[1]
    vm.store_memory(addr, len(vm.input_buffer))
    for i, c in enumerate(vm.input_buffer, start=1):
        vm.store_memory(addr + i, ord(c))


def tiger_tstrcmp_reg(vm):
    # This could be implemented in pure HERA but I'm too lazy.
    s1 = vm.registers[1]
    s2 = vm.registers[2]
    n1 = vm.load_memory(s1)
    n2 = vm.load_memory(s2)
    i = 0
    while i < n1 and i < n2:
        c1 = vm.load_memory(s1 + i + 1)
        c2 = vm.load_memory(s2 + i + 1)
        if c1 < c2:
            vm.registers[1] = to_u16(-1)
            return
        elif c1 > c2:
            vm.registers[1] = 1
            return
        i += 1

    if n1 == n2:
        vm.registers[1] = 0
    elif n1 < n2:
        vm.registers[1] = to_u16(-1)
    else:
        vm.registers[1] = 1


# The standard library with parameters-in-registers functions.
TIGER_STDLIB_REG = """
LABEL(printint)
  __eval("stdlib.tiger_printint_reg(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(print)
  __eval("stdlib.tiger_print_reg(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(println)
  __eval("stdlib.tiger_println_reg(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(div)
  __eval("stdlib.tiger_div_reg(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(mod)
  __eval("stdlib.tiger_mod_reg(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(getchar_ord)
  __eval("stdlib.tiger_getchar_ord_reg(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(putchar_ord)
  __eval("stdlib.tiger_putchar_ord_reg(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(flush)
  __eval("stdlib.tiger_flush_reg(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(printbool)
  __eval("stdlib.tiger_printbool_reg(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(ungetchar)
  __eval("stdlib.tiger_ungetchar_reg(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(tstrcmp)
  __eval("stdlib.tiger_tstrcmp_reg(vm)")
  RETURN(FP_alt, PC_ret)


LABEL(exit)
  HALT()


LABEL(getline)
  INC(SP, 1)
  STORE(PC_ret, 0, FP)
  __eval("stdlib.tiger_getline_preamble_reg(vm)")
  MOVE(R12, SP)

  MOVE(FP_alt, SP)
  // R1 was set to the length of the string by tiger_getline_preamble_stack
  CALL(FP_alt, malloc)

  __eval("stdlib.tiger_getline_epilogue_reg(vm)")
  LOAD(PC_ret, 0, FP)
  DEC(SP, 1)

  RETURN(FP_alt, PC_ret)


// fail if wrapped past 0xffff, hard-coded, but check
// ** malloc(n_cells: int) : int  // returns address of allocated space
LABEL(malloc)
     // Reg 1  = argument 1 (number of cells needed)
     // OK to just leave PC_ret as the return address, since we have no calls (except exiting)
     SET(R11,first_space_for_fsheap)  // Ok to use R11, aka Rt, since no "not" pseudo-op or register-branches
     LOAD(R10,0,R11)	// R10 is address of first free
     FLAGS(R10)
   BNZR(fsheap_already_initialized)
     MOVE(R10, R11)
     INC(R10, 1)      // initialize fs heap to its first free element then allocate from there
     STORE(R10,0,R11) // defensive programming --- this should be overwritten later and thus dead
   LABEL(fsheap_already_initialized)

     MOVE(R9, R10)	// Hold on to value to be returned
     ADD(R10,R10,R1)	// ALLOCATE -- FAIL BELOW IF WRAPPED PAST END OF ADDRESS SPACE OR "last_space_for_fsheap"
     BCR(malloc_give_up)
     SET(R1,last_space_for_fsheap)
     CMP(R10, R1)	// want (unsigned) R10 < R1, so R10-R1 should BORROW, i.e. NOT carry!
     BCR(malloc_give_up)
     STORE(R10,0,R11)   // STORE BACK THE NEXT FREE SPACE

     MOVE(R1,R9)	// Value to be returned was in R9
     RETURN(FP_alt, PC_ret)	// Normal return from malloc

   LABEL(malloc_inconsistent)
     SET(R1, malloc_inconsistent_error)
     BRR(malloc_exit)
   LABEL(malloc_give_up)
     SET(R1,malloc_out_of_memory_error)
   LABEL(malloc_exit)
     CALL(FP_alt,print)
     SET(R1,127)
     CALL(FP_alt,exit)
     HALT() // shouldn't ever get here


LABEL(not)
	// Reg 1 <-- argument
     CMP(R1,R0)
     BZR(tstdlib_label_arg_was_false)
     SET(R1,0)
     BR(tstdlib_label_return_from_not)
     LABEL(tstdlib_label_arg_was_false)
     SET(R1,1)
     LABEL(tstdlib_label_return_from_not)
     RETURN(FP_alt, PC_ret)


//    size(s:string) : int
LABEL(size)
     // Reg 1 <-- argument (address of string)
     LOAD(R1,0,R1)  // Reg 1 now has the size of string
     RETURN(FP_alt, PC_ret)


//   ord(s:string) : int
LABEL(ord)
     // Reg 1 <-- argument (address of string)
     LOAD(R1,1,R1)  // Reg 1 now has the 1st character of the string
     RETURN(FP_alt, PC_ret)


LABEL(getchar)
  INC(SP, 1)

  SET(R1, 2)
  MOVE(FP_alt, SP)
  CALL(FP_alt, malloc)

  // R1 = pointer to newly-allocated string

  // Store size of new string in M[R1]
  SET(Rt, 1)
  STORE(Rt, 0, R1)

  // Save R1 on the stack
  STORE(R1, 0, FP)

  CALL(FP_alt, getchar_ord)

  // R1 = ordinal number of character read

  // Load string pointer from stack
  LOAD(Rt, 0, FP)

  // Insert character into string
  STORE(R1, 1, Rt)

  // Set R1 to return value (pointer to string)
  MOVE(R1, Rt)

  DEC(SP, 1)
"""


# The data segment for the parameters-in-registers functions.
TIGER_STDLIB_REG_DATA = (
    TIGER_STDLIB_STACK_DATA
    + """\
DLABEL(TIGER_STDLIB_REG_UNIMPLEMENTED)
TIGER_STRING("Error -- call to unimplemented function from Tiger-stdlib-reg.hera\n")
"""
)



"""
Convenient interface for parsing, type-checking and optionally preprocessing a HERA
file.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: March 2019
"""
from .checker import check
from .data import Program, Settings
from .parser import parse
from .utils import handle_messages, read_file_or_stdin


def load_program(text: str, settings=Settings()) -> Program:
    """
    Parse the string into a program, type-check it, and preprocess it. A tuple
    (ops, symbol_table) is returned.

    The return value of this function is valid input to the VirtualMachine.run method.
    """
    oplist, parse_messages = parse(text, settings=settings)
    program, check_messages = check(oplist, settings=settings)
    handle_messages(settings, parse_messages.extend(check_messages))
    return program


def load_program_from_file(path: str, settings=Settings()) -> Program:
    """
    Convenience function to a read a file and then invoke `load_program_from_str` on its
    contents.
    """
    text = read_file_or_stdin(path, settings)
    if path == "-":
        path = "<stdin>"

    oplist, parse_messages = parse(text, path=path, settings=settings)
    program, check_messages = check(oplist, settings=settings)
    handle_messages(settings, parse_messages.extend(check_messages))
    return program



"""
Debug HERA programs.

The code in this module makes an important distinction between "real ops" and "original
ops." Original ops are the HERA operations as they appear in the program that the user
wrote. Real ops are the original ops transformed by the preprocessor into something that
the virtual machine can actually run. For example, a single original SET op corresponds
to two real ops, SETLO and SETHI.

Internally, the debugger operates on real ops, but whenever it displays output to the
user, it must be in terms of original ops.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: February 2019
"""
import copy
from typing import Dict, List, Optional

try:
    import readline  # noqa: F401
except ImportError:
    pass

from hera.data import Label, Program, Settings
from hera.op import AbstractOperation
from hera.vm import VirtualMachine


class Debugger:
    """
    A class for debugging. External users should generally use the `debug` function
    in hera.debugger.shell instead of this class.
    """

    def __init__(self, program: Program, settings: Settings) -> None:
        self.settings = settings
        self.program = program
        self.symbol_table = program.symbol_table
        # A map from instruction numbers (i.e., possible values of the program counter)
        # to human-readable line numbers.
        self.breakpoints = {}  # type: Dict[int, str]
        self.vm = VirtualMachine()
        # How many CALLs without RETURNs?
        self.calls = 0
        # Back-up of the debugger's state, to implement the "undo" command.
        # Set to None when no operations have been performed, and to "undone" when an
        # "undo" has just been executed.
        self.old = None  # type: Optional[Debugger]

        for data_op in self.program.data:
            data_op.execute(self.vm)

    def save(self) -> None:
        self.old = copy.copy(self)
        self.old.symbol_table = self.program.symbol_table.copy()
        self.old.breakpoints = self.breakpoints.copy()
        self.old.vm = self.vm.copy()

    def get_breakpoints(self) -> Dict[int, str]:
        return self.breakpoints

    def set_breakpoint(self, b: int) -> None:
        self.breakpoints[b] = self.get_breakpoint_name(b)

    def at_breakpoint(self) -> bool:
        return not self.finished() and self.vm.pc in self.breakpoints

    def next(self, *, step) -> None:
        if self.finished():
            return

        if not step and self.program.code[self.vm.pc].original.name == "CALL":
            calls = self.calls
            # step=True prevents infinite regress.
            self.next(step=True)
            while (
                not self.finished() and not self.at_breakpoint() and self.calls > calls
            ):
                self.next(step=True)
        else:
            for real_op in self.real_ops():
                if real_op.name == "CALL":
                    self.calls += 1
                elif real_op.name == "RETURN":
                    self.calls -= 1

                real_op.execute(self.vm)

    def reset(self) -> None:
        self.vm.reset()

    def op(self, index=None) -> AbstractOperation:
        """
        Return the original operation at the given index, which defaults to the
        current program counter.
        """
        if index is None:
            index = self.vm.pc
        return self.program.code[index].original

    def real_ops(self) -> List[AbstractOperation]:
        """
        Return all the real ops that correspond to the current original op. See
        module docstring for explanation of terminology.
        """
        original = self.op()
        end = self.vm.pc
        while end < len(self.program.code) and self.op(end) == original:
            end += 1

        return self.program.code[self.vm.pc : end]

    def resolve_location(self, b: str) -> int:
        """Resolve a user-supplied location string into an instruction number"""
        if b == ".":
            return self.vm.pc

        if ":" in b:
            path, lineno = b.split(":", maxsplit=1)
        else:
            path = self.op().loc.path
            lineno = b

        try:
            lineno_as_int = int(lineno)
        except ValueError:
            try:
                opno = self.program.symbol_table[b]
                assert isinstance(opno, Label)
                return opno
            except (KeyError, AssertionError):
                raise ValueError("could not locate label `{}`.".format(b)) from None
        else:
            for pc, op in enumerate(self.program.code):
                if op.loc.path == path and op.loc.line == lineno_as_int:
                    return pc

            raise ValueError("could not find corresponding line.")

    def get_breakpoint_name(self, b: int, *, append_label=True) -> str:
        """
        Turn an instruction number into a human-readable location string with the file
        path and line number. More or less the inverse of `resolve_location`.
        """
        op = self.program.code[b].original or self.program.code[b]
        path = "<stdin>" if op.loc.path == "-" else op.loc.path
        loc = path + ":" + str(op.loc.line)

        if append_label:
            # Look for a label corresponding to the breakpoint.
            label = self.find_label(b)
            if label is not None:
                return "{} ({})".format(loc, label)

        return loc

    def find_label(self, ino: int) -> Optional[str]:
        """Find a label, if one exists, corresponding to the instruction number."""
        for symbol, value in self.program.symbol_table.items():
            if value == ino and isinstance(value, Label):
                return symbol
        return None

    def finished(self) -> bool:
        return self.vm.halted or self.vm.pc >= len(self.program.code)

    def empty(self) -> bool:
        return len(self.program.code) == 0


def reverse_lookup_label(symbol_table: Dict[str, int], value: int) -> Optional[str]:
    """
    Return the name of the label that maps to `value`, or None if no such label is
    found. Constants and data labels are ignored.
    """
    for k, v in symbol_table.items():
        if value == v and isinstance(v, Label):
            return k
    return None



from hera.data import DEFAULT_DATA_START
from .utils import execute_program_helper


def test_addition_program(capsys):
    program = """\
SET(R1, 20)
SET(R2, 22)
ADD(R3, R1, R2)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 20
    assert vm.registers[2] == 22
    assert vm.registers[3] == 42

    for r in vm.registers[4:]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block

    for x in vm.memory:
        assert x == 0

    assert "Warning" not in capsys.readouterr().err


def test_loop_program(capsys):
    program = """\
SET(R1, 1)
SET(R2, 10)
LABEL(top)
INC(R1, 1)
CMP(R1, R2)
BZ(bottom)
BR(top)
LABEL(bottom)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 10
    assert vm.registers[2] == 10

    for r in vm.registers[3:10]:
        assert r == 0

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block

    for x in vm.memory:
        assert x == 0

    assert "Warning" not in capsys.readouterr().err


def test_function_call_program(capsys):
    program = """\
SET(R1, 8)
CALL(FP_alt, times_two)
HALT()

// Multiply R1 by two, in-place
LABEL(times_two)
  LSL(R1, R1)
  RETURN(FP_alt, PC_ret)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 16

    for r in vm.registers[2:10]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block

    for x in vm.memory:
        assert x == 0

    assert "Warning" not in capsys.readouterr().err


def test_fibonacci_program(capsys):
    program = """\
/*

fib(n)
  i = 1
  fib_i = 1
  fib_i_minus_1 = 0
  while i < n
    tmp = fib_i
    fib_i = fib_i + fib_i_minus_1
    fib_i_minus_1 = fib_i
    i += 1
  return fib_i

*/

CBON()

// n = 12
SET(R1, 12)

// fib_i = 1
SET(R2, 1)
// fib_i_minus_1 = 0
SET(R3, 0)
// i = 1
SET(R4, 1)

LABEL(top)
CMP(R4, R1)
BZ(bottom)

// tmp = fib_i
MOVE(R5, R2)
ADD(R2, R2, R3)
MOVE(R3, R5)
INC(R4, 1)

BR(top)
LABEL(bottom)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 12
    assert vm.registers[2] == 144
    assert vm.registers[3] == 89
    assert vm.registers[4] == 12
    assert vm.registers[5] == 89

    for r in vm.registers[6:10]:
        assert r == 0

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block

    for x in vm.memory:
        assert x == 0

    assert "Warning" not in capsys.readouterr().err


def test_data_easy_program(capsys):
    program = """\
DLABEL(X)
INTEGER(42)

SET(R1, X)
LOAD(R2, 0, R1)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == DEFAULT_DATA_START
    assert vm.registers[2] == 42

    for r in vm.registers[3:]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block

    assert vm.memory[DEFAULT_DATA_START] == 42

    assert "Warning" not in capsys.readouterr().err


def test_dskip_program(capsys):
    program = """\
DLABEL(first_array)
INTEGER(42)
DSKIP(10)
INTEGER(84)

SET(R1, first_array)
LOAD(R2, 0, R1)
LOAD(R3, 11, R1)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == DEFAULT_DATA_START
    assert vm.registers[2] == 42
    assert vm.registers[3] == 84

    for r in vm.registers[4:]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block

    assert vm.memory[DEFAULT_DATA_START] == 42
    assert vm.memory[DEFAULT_DATA_START + 11] == 84

    assert "Warning" not in capsys.readouterr().err


def test_loop_and_constant_program(capsys):
    program = """\
CONSTANT(N, 100)

SET(R1, N)
SET(R2, 0)
SET(R3, 0)

LABEL(top)
CMP(R1, R2)
BZ(bottom)
ADD(R3, R3, R2)
INC(R2, 1)
BR(top)
LABEL(bottom)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 100
    assert vm.registers[2] == 100
    assert vm.registers[3] == 5050

    for r in vm.registers[4:10]:
        assert r == 0

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block

    assert "Warning" not in capsys.readouterr().err


def test_hera_boilerplate_program():
    program = """\
#include <HERA.h>

void HERA_main() {
   SET(R1, 42);
   SET(R2, 42)
}
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 42
    assert vm.registers[2] == 42

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_relative_branching():
    program = """\
SET(R1, 10)
SET(R2, 32)
BRR(after)
SET(R2, 656)
LABEL(after)
ADD(R3, R1, R2)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 10
    assert vm.registers[2] == 32
    assert vm.registers[3] == 42


def test_relative_branching_backwards():
    program = """\
BRR(after)
LABEL(before)
HALT()
LABEL(after)
SET(R1, 42)
BRR(before)
SET(R1, 666)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 42


def test_relative_branching_in_large_program():
    program = ("NOP()\n") * 255 + "BRR(next)\nSET(R2, 666)\nLABEL(next)\nSET(R1, 42)"
    vm = execute_program_helper(program)

    assert vm.registers[1] == 42
    assert vm.registers[2] == 0


def test_branching_by_register():
    program = """\
SET(R1, end)
SET(R2, 42)
BR(R1)

SET(R2, 666)
LABEL(end)
SET(R3, 84)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 7
    assert vm.registers[2] == 42
    assert vm.registers[3] == 84


def test_no_extraneous_output(capsys):
    execute_program_helper("SET(R1, 42)")

    captured = capsys.readouterr()
    assert captured.out == ""
    assert (
        captured.err
        == """\


Virtual machine state after execution:
    R1  = 0x002a = 42 = '*'

    All flags are OFF
"""
    )


def test_two_TIGER_STRING_ops():
    program = """\
DLABEL(s1) TIGER_STRING("hello")
DLABEL(s2) TIGER_STRING("world")

SET(R1, s1)
SET(R2, s2)
    """
    vm = execute_program_helper(program)

    s1_addr = vm.registers[1]
    s2_addr = vm.registers[2]
    assert vm.load_memory(s1_addr) == 5
    assert vm.load_memory(s2_addr) == 5

    for i in range(5):
        assert vm.load_memory(s1_addr + i + 1) == ord("hello"[i])
        assert vm.load_memory(s2_addr + i + 1) == ord("world"[i])


def test_some_neglected_flag_ops():
    program = """\
CMP(R0, R0)
BLER(after_bler)
SET(R1, 666)
LABEL(after_bler)

SET(Rt, 1)
CMP(Rt, R0)
BGR(after_bgr)
SET(R2, 666)
LABEL(after_bgr)

CMP(R0, R0)
BULER(after_buler)
SET(R3, 666)
LABEL(after_buler)

SET(Rt, 1)
CMP(Rt, R0)
BUGR(after_bugr)
SET(R4, 666)
LABEL(after_bugr)

SET(Rt, 1)
FLAGS(Rt)
BNZR(after_bnzr)
SET(R5, 666)
LABEL(after_bnzr)

FLAGS(R0)
BNS(after_bns)
SET(R6, 666)
LABEL(after_bns)

FLAGS(R0)
BNSR(after_bnsr)
SET(R7, 666)
LABEL(after_bnsr)

FON(0b100)
BVR(after_bvr)
SET(R8, 666)
LABEL(after_bvr)

FOFF(0b100)
BNV(after_bnv)
SET(R9, 666)
LABEL(after_bnv)

FOFF(0b100)
BNVR(after_bnvr)
SET(R10, 666)
LABEL(after_bnvr)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 0
    assert vm.registers[2] == 0
    assert vm.registers[3] == 0
    assert vm.registers[4] == 0
    assert vm.registers[5] == 0
    assert vm.registers[6] == 0
    assert vm.registers[7] == 0
    assert vm.registers[8] == 0
    assert vm.registers[9] == 0
    assert vm.registers[10] == 0


def test_OPCODE_instructions():
    program = """\
// SETLO(R1, 0xFF)
OPCODE(0xe1ff)
// SETHI(R3, 42)
OPCODE(0xf32a)
// INC(R10, 1)
OPCODE(0x3a80)
// DEC(R1, 20)
OPCODE(0x31d3)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 65515
    assert vm.registers[3] == 10752
    assert vm.registers[10] == 1



import os
from setuptools import find_packages, setup


dpath = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(dpath, "README.md"), "r") as f:
    long_description = f.read()


setup(
    name="hera-py",
    version="0.8.1",
    description=(
        "Interpreter for the Haverford Educational RISC Architecture (HERA) "
        "assembly language"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    author="Ian Fisher",
    author_email="iafisher@protonmail.com",
    entry_points={"console_scripts": ["hera = hera.main:external_main"]},
    install_requires=["typing==3.6.6"],  # typing module is necessary for Python 3.4
    packages=find_packages(exclude=["tests"]),
    project_urls={"Source": "https://github.com/iafisher/hera-py"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Topic :: Software Development :: Assemblers",
    ],
)



import pytest

from hera.checker import get_labels, operation_length, substitute_label, typecheck
from hera.data import Settings, Token
from hera.op import (
    ADD,
    BNZ,
    CALL,
    CMP,
    FLAGS,
    INC,
    MOVE,
    NEG,
    NOT,
    SET,
    SETHI,
    SETLO,
    SETRF,
)
from hera.parser import parse


@pytest.fixture
def settings():
    return Settings()


def helper(opstr, symbol_table={}):
    ops, messages = parse(opstr)
    if not ops or messages.errors:
        return messages.errors
    else:
        return ops[0].typecheck(symbol_table).errors


def valid(opstr, symbol_table={}):
    assert len(helper(opstr, symbol_table)) == 0


def invalid(opstr, msg, symbol_table={}):
    errors = helper(opstr, symbol_table)

    assert len(errors) > 0
    assert msg in errors[0][0]


def test_typecheck_SET():
    valid("SET(R1, 42)")
    valid("SET(R1, 0xFFFF)")
    valid("SET(R1, -0x7FFF)")


def test_typecheck_SETLO():
    valid("SETLO(R2, 42)")
    valid("SETLO(R2, 0xFF)")
    valid("SETLO(R2, -0x7F)")


def test_typecheck_SETHI():
    valid("SETHI(R2, 42)")
    valid("SETHI(R2, 0xFF)")
    valid("SETHI(R2, -0x7F)")


def test_typecheck_AND():
    valid("AND(R3, R4, R5)")


def test_typecheck_OR():
    valid("OR(R3, R4, R5)")


def test_typecheck_ADD():
    valid("ADD(R3, R4, R5)")


def test_typecheck_SUB():
    valid("SUB(R3, R4, R5)")


def test_typecheck_MUL():
    valid("MUL(R3, R4, R5)")


def test_typecheck_XOR():
    valid("XOR(R3, R4, R5)")


def test_typecheck_INC():
    valid("INC(R6, 42)")
    valid("INC(R6, 1)")
    valid("INC(R6, 64)")


def test_typecheck_DEC():
    valid("DEC(R6, 42)")
    valid("DEC(R6, 1)")
    valid("DEC(R6, 64)")


def test_typecheck_LSL():
    valid("LSL(R7, R8)")


def test_typecheck_LSR():
    valid("LSR(R7, R8)")


def test_typecheck_LSL8():
    valid("LSL8(R7, R8)")


def test_typecheck_LSR8():
    valid("LSR8(R7, R8)")


def test_typecheck_ASL():
    valid("ASL(R7, R8)")


def test_typecheck_ASR():
    valid("ASR(R7, R8)")


def test_typecheck_SAVEF():
    valid("SAVEF(R9)")


def test_typecheck_RSTRF():
    valid("RSTRF(R9)")


def test_typecheck_FON():
    valid("FON(0b10101)")
    valid("FON(0b01000)")
    valid("FON(0b11111)")
    valid("FON(0)")


def test_typecheck_FOFF():
    valid("FOFF(0b10101)")
    valid("FOFF(0b01000)")
    valid("FOFF(0b11111)")
    valid("FOFF(0)")


def test_typecheck_FSET5():
    valid("FSET5(0b10101)")
    valid("FSET5(0b01000)")
    valid("FSET5(0b11111)")
    valid("FSET5(0)")


def test_typecheck_FSET4():
    valid("FSET4(0b1010)")
    valid("FSET4(0b0100)")
    valid("FSET4(0b1111)")
    valid("FSET4(0)")


def test_typecheck_LOAD():
    valid("LOAD(R1, 0, R2)")
    valid("LOAD(R1, 0b11111, R2)")


def test_typecheck_STORE():
    valid("STORE(R1, 0, R2)")
    valid("STORE(R1, 0b11111, R2)")


def test_typecheck_CALL():
    valid("CALL(R12, R11)")
    valid("CALL(R12, f)", {"f": 0})


def test_typecheck_RETURN():
    valid("RETURN(R12, R13)")
    invalid("RETURN(R12, f)", "expected register", {"f": 0})


def test_typecheck_BR():
    valid("BR(R11)")
    valid("BR(l)", {"l": 0})


def test_typecheck_BRR():
    valid("BRR(0xFF)")
    valid("BRR(-0x7F)")


def test_typecheck_BL():
    valid("BL(R11)")
    valid("BL(l)", {"l": 0})


def test_typecheck_BLR():
    valid("BLR(0xFF)")
    valid("BLR(-0x7F)")


def test_typecheck_BGE():
    valid("BGE(R11)")
    valid("BGE(l)", {"l": 0})


def test_typecheck_BGER():
    valid("BGER(0xFF)")
    valid("BGER(-0x7F)")


def test_typecheck_BLE():
    valid("BLE(R11)")
    valid("BLE(l)", {"l": 0})


def test_typecheck_BLER():
    valid("BLER(0xFF)")
    valid("BLER(-0x7F)")


def test_typecheck_BG():
    valid("BG(R11)")
    valid("BG(l)", {"l": 0})


def test_typecheck_BGR():
    valid("BGR(0xFF)")
    valid("BGR(-0x7F)")


def test_typecheck_BULE():
    valid("BULE(R11)")
    valid("BULE(l)", {"l": 0})


def test_typecheck_BULER():
    valid("BULER(0xFF)")
    valid("BULER(-0x7F)")


def test_typecheck_BUG():
    valid("BUG(R11)")
    valid("BUG(l)", {"l": 0})


def test_typecheck_BUGR():
    valid("BUGR(0xFF)")
    valid("BUGR(-0x7F)")


def test_typecheck_BZ():
    valid("BZ(R11)")
    valid("BZ(l)", {"l": 0})


def test_typecheck_BZR():
    valid("BZR(0xFF)")
    valid("BZR(-0x7F)")


def test_typecheck_BNZ():
    valid("BNZ(R11)")
    valid("BNZ(l)", {"l": 0})


def test_typecheck_BNZR():
    valid("BNZR(0xFF)")
    valid("BNZR(-0x7F)")


def test_typecheck_BC():
    valid("BC(R11)")
    valid("BC(l)", {"l": 0})


def test_typecheck_BCR():
    valid("BCR(0xFF)")
    valid("BCR(-0x7F)")


def test_typecheck_BNC():
    valid("BNC(R11)")
    valid("BNC(l)", {"l": 0})


def test_typecheck_BNCR():
    valid("BNCR(0xFF)")
    valid("BNCR(-0x7F)")


def test_typecheck_BS():
    valid("BS(R11)")
    valid("BS(l)", {"l": 0})


def test_typecheck_BSR():
    valid("BSR(0xFF)")
    valid("BSR(-0x7F)")


def test_typecheck_BNS():
    valid("BNS(R11)")
    valid("BNS(l)", {"l": 0})


def test_typecheck_BNSR():
    valid("BNSR(0xFF)")
    valid("BNSR(-0x7F)")


def test_typecheck_BV():
    valid("BV(R11)")
    valid("BV(l)", {"l": 0})


def test_typecheck_BVR():
    valid("BVR(0xFF)")
    valid("BVR(-0x7F)")


def test_typecheck_BNV():
    valid("BNV(R11)")
    valid("BNV(l)", {"l": 0})


def test_typecheck_BNVR():
    valid("BNVR(0xFF)")
    valid("BNVR(-0x7F)")


def test_typecheck_SETRF():
    valid("SETRF(R1, 42)")
    valid("SETRF(R1, 0xFFFF)")
    valid("SETRF(R1, -0x7FFF)")


def test_typecheck_MOVE():
    valid("MOVE(R1, R2)")


def test_typecheck_CMP():
    valid("CMP(R1, R2)")


def test_typecheck_NEG():
    valid("NEG(R1, R2)")


def test_typecheck_NOT():
    valid("NOT(R1, R2)")


def test_typecheck_CBON():
    valid("CBON()")


def test_typecheck_CON():
    valid("CON()")


def test_typecheck_COFF():
    valid("COFF()")


def test_typecheck_CCBOFF():
    valid("CCBOFF()")


def test_typecheck_FLAGS():
    valid("FLAGS(R1)")


def test_typecheck_NOP():
    valid("NOP()")


def test_typecheck_HALT():
    valid("HALT()")


def test_typecheck_LABEL():
    valid("LABEL(l)")


def test_typecheck_CONSTANT():
    valid("CONSTANT(N, 0xFFFF)")
    valid("CONSTANT(N, -0x7FFF)")


def test_typecheck_DLABEL():
    valid("DLABEL(l)")


def test_typecheck_INTEGER():
    valid("INTEGER(0xFFFF)")
    valid("INTEGER(-0x7FFF)")


def test_typecheck_LP_STRING():
    valid('LP_STRING("hello!")')


def test_typecheck_TIGER_STRING():
    valid('TIGER_STRING("hello!")')


def test_typecheck_DSKIP():
    valid("DSKIP(0xFFFF)")
    valid("DSKIP(0)")


def test_typecheck_SWI():
    valid("SWI(0b1111)")
    valid("SWI(0b0110)")
    valid("SWI(0)")


def test_typecheck_RTI():
    valid("RTI()")


def test_typecheck_OPCODE():
    # AND(R1, R7, R12)
    valid("OPCODE(0x817c)")
    # BRR(0)
    valid("OPCODE(0)")
    # A branch instruction that doesn't exist.
    invalid("OPCODE(0b0001000100000101)", "not a HERA instruction")


def test_typecheck_print_reg():
    valid("print_reg(R1)")


def test_typecheck_print():
    valid('print("hello, world!")')


def test_typecheck_println():
    valid('println("hello, world!")')


def test_typecheck___eval():
    valid("__eval(\"print('hello')\")")


def test_typecheck_undefined_symbol():
    invalid("SET(R1, N)", "undefined constant")


def test_typecheck_unknown_instruction():
    invalid("IF(R1)", "unknown instruction")


def test_typecheck_unknown_branch_instruction():
    invalid("BNWR(R1)", "unknown instruction")


def test_typecheck_single_error():
    # Second argument to SETHI is out of range.
    program = [SETLO(Token.R(1), Token.Int(10)), SETHI(Token.R(1), Token.Int(1000))]
    symbol_table, messages = typecheck(program)

    assert len(messages.errors) == 1
    assert "integer must be in range [-128, 256)" in messages.errors[0][0]


def test_typecheck_multiple_errors():
    program = [
        ADD(Token.R(1), Token.Int(10)),
        INC(Token.R(3), Token.Int(1), Token.Int(2)),
    ]
    symbol_table, messages = typecheck(program)

    assert len(messages.errors) == 3

    assert "ADD" in messages.errors[0][0]
    assert "too few" in messages.errors[0][0]

    assert "expected register" in messages.errors[1][0]

    assert "INC" in messages.errors[2][0]
    assert "too many" in messages.errors[2][0]


def test_operation_length_of_register_branch_with_label():
    assert operation_length(BNZ(Token.Sym("l"))) == 3


def test_operation_length_of_register_branch_with_register():
    assert operation_length(BNZ(Token.R(1))) == 1


def test_operation_length_of_SET():
    assert operation_length(SET(Token.R(1), Token.Int(10))) == 2


def test_operation_length_of_SETRF():
    assert operation_length(SETRF(Token.R(1), Token.Int(10))) == 4


def test_operation_length_of_MOVE():
    assert operation_length(MOVE(Token.R(1), Token.R(2))) == 1


def test_operation_length_of_CMP():
    assert operation_length(CMP(Token.R(1), Token.R(0))) == 2


def test_operation_length_of_NEG():
    assert operation_length(NEG(Token.R(7), Token.R(15))) == 2


def test_operation_length_of_NOT():
    assert operation_length(NOT(Token.R(5), Token.R(7))) == 3


def test_operation_lentgh_of_FLAGS():
    assert operation_length(FLAGS(Token.R(3))) == 2


def test_operation_length_of_CALL_with_label():
    assert operation_length(CALL(Token.R(12), Token.Sym("l"))) == 3


def test_operation_length_of_CALL_with_register():
    assert operation_length(CALL(Token.R(12), Token.R(13))) == 1


def test_get_labels_with_invalid_code(settings):
    labels, messages = get_labels([CALL(Token.Sym("l"))], settings)

    assert len(labels) == 0
    assert len(messages.errors) == 0


def test_substitute_label_with_SETLO():
    program = SETLO(Token.R(1), Token.Sym("N"))
    labels = {"N": 10}
    assert substitute_label(program, labels) == SETLO(Token.R(1), Token.Int(10))


def test_substitute_label_with_SETHI():
    program = SETHI(Token.R(1), Token.Sym("N"))
    labels = {"N": 10}
    assert substitute_label(program, labels) == SETHI(Token.R(1), Token.Int(10))


def test_substitute_label_with_other_op():
    program = INC(Token.R(1), Token.Sym("N"))
    labels = {"N": 10}
    assert substitute_label(program, labels) == INC(Token.R(1), Token.Int(10))
